"""
Generic framework path manipulation
"""

__all__ = ["infoForFramework"]

# This regexp should find:
#   \1 - framework location
#   \2 - framework name
#   \3 - framework version (optional)
#
FRAMEWORK_RE_STR = r"""(^.*)(?:^|/)(\w+).framework(?:/(?:Versions/([^/]+)/)?\2)?$"""
FRAMEWORK_RE = None


def infoForFramework(filename):
    """returns (location, name, version) or None"""
    global FRAMEWORK_RE
    if FRAMEWORK_RE is None:
        import re

        FRAMEWORK_RE = re.compile(FRAMEWORK_RE_STR)
    is_framework = FRAMEWORK_RE.findall(filename)
    if not is_framework:
        return None
    return is_framework[-1]



import collections.abc
import datetime

from objc import _objc

__all__ = [
    "registerListType",
    "registerMappingType",
    "registerSetType",
    "registerDateType",
]


def registerListType(type_object):
    """
    Register 'type' as a list-like type that will be proxied
    as an NSMutableArray subclass.
    """
    if _objc.options._sequence_types is None:
        _objc.options._sequence_types = ()

    _objc.options._sequence_types += (type_object,)


def registerMappingType(type_object):
    """
    Register 'type' as a dictionary-like type that will be proxied
    as an NSMutableDictionary subclass.
    """
    if _objc.options._mapping_types is None:
        _objc.options._mapping_types = ()

    _objc.options._mapping_types += (type_object,)


def registerSetType(type_object):
    """
    Register 'type' as a set-like type that will be proxied
    as an NSMutableSet subclass.
    """
    if _objc.options._set_types is None:
        _objc.options._set_types = ()

    _objc.options._set_types += (type_object,)


def registerDateType(type_object):
    """
    Register 'type' as a date-like type that will be proxied
    as an NSDate subclass.
    """
    if _objc.options._date_types is None:
        _objc.options._date_types = ()

    _objc.options._date_types += (type_object,)


registerListType(collections.abc.Sequence)
registerListType(range)
registerMappingType(collections.abc.Mapping)
registerMappingType(dict)
registerSetType(set)
registerSetType(frozenset)
registerSetType(collections.abc.Set)
registerDateType(datetime.date)
registerDateType(datetime.datetime)



"""
A number of Cocoa API's have a 'context' argument that is a plain 'void*'
in ObjC, and an Integer value in Python.  The 'context' object defined here
allows you to get a unique integer number that can be used as the context
argument for any Python object, and retrieve that object later on using the
context number.

Usage::

    ...
    ctx = objc.context.register(myContext)
    someObject.observeValueForKeyPath_ofObject_change_context_(
        kp, obj, {}, ctx)
    ...

and in the callback::

    def observeValueForKeyPath_ofObject_change_context_(self,
        kp, obj, change, ctx):

        myContext = objc.context.get(ctx)
        ...

Use ``objc.context.unregister`` to remove the registration of ``myObject``
when you're done. The argument to unregister is the same object as was
passed in during registration.
"""
__all__ = ("context",)
from typing import Any


class ContextRegistry(object):
    def __init__(self):
        self._registry = {}

    def register(self, value: Any) -> int:
        uniq = id(value)
        self._registry[uniq] = value
        return uniq

    def unregister(self, value: Any):
        try:
            del self._registry[id(value)]
        except KeyError:
            pass

    def get(self, uniq) -> Any:
        return self._registry[uniq]


context = ContextRegistry()



import warnings

from objc import options as _options


def setVerbose(value):
    warnings.warn("Set objc.options.verbose instead", DeprecationWarning)
    _options.verbose = bool(value)


def getVerbose():
    warnings.warn("Read objc.options.verbose instead", DeprecationWarning)
    return _options.verbose


def setUseKVOForSetattr(value):
    warnings.warn("Set objc.options.use_kvo instead", DeprecationWarning, 2)
    _options.use_kvo = bool(value)


def getUseKVOForSetattr():
    warnings.warn("Read objc.options.use_kvo instead", DeprecationWarning)
    return _options.use_kvo



__all__ = (
    "object_property",
    "bool_property",
    "array_property",
    "set_property",
    "dict_property",
)

import sys
from copy import copy as copy_func

from objc import (
    _C_BOOL,
    _C_ID,
    _C_NSBOOL,
    NULL,
    _C_NSUInteger,
    ivar,
    lookUpClass,
    selector,
)

import collections.abc

NSSet = lookUpClass("NSSet")
NSObject = lookUpClass("NSObject")


def _str(value):
    return value.decode("ascii")


def attrsetter(prop, name, copy):
    if copy:

        def func(self, value):
            if isinstance(value, NSObject):
                setattr(self, name, value.copy())
            else:
                setattr(self, name, copy_func(value))

    else:

        def func(self, value):
            setattr(self, name, value)

    return func


def attrgetter(name):
    def func(self):
        return getattr(self, name)

    return func


def _return_value(value):
    def func(self):
        return value

    return func


def _dynamic_getter(name):
    def getter(an_object):
        m = getattr(an_object.pyobjc_instanceMethods, name)
        return m()

    getter.__name__ = name
    return getter


def _dynamic_setter(name):
    def setter(an_object, value):
        m = getattr(an_object.pyobjc_instanceMethods, name.replace(":", "_"))
        return m(value)

    setter.__name__ = name
    return setter


class object_property:
    def __init__(
        self,
        name=None,
        read_only=False,
        copy=False,
        dynamic=False,
        ivar=None,
        typestr=_C_ID,
        depends_on=None,
    ):
        self.__created = False
        self.__inherit = False
        self._name = name
        self._typestr = typestr
        self._ro = read_only
        self._copy = copy
        self._dynamic = dynamic
        self._ivar = ivar
        self._getter = None
        self._setter = None
        self._validate = None
        if depends_on is None:
            self._depends_on = None
        else:
            self._depends_on = set(depends_on)

        self.__getprop = None
        self.__setprop = None
        self.__parent = None

    def _clone(self):
        if self._depends_on is None:
            depends = None
        else:
            depends = self._depends_on.copy()

        v = type(self)(
            name=self._name,
            read_only=self._ro,
            copy=self._copy,
            dynamic=self._dynamic,
            ivar=self._ivar,
            typestr=self._typestr,
            depends_on=depends,
        )
        v.__inherit = True

        v.__getprop = self.__getprop
        v.__setprop = self.__setprop
        v.__parent = self

        return v

    def __pyobjc_class_setup__(self, name, class_dict, instance_methods, class_methods):
        self.__created = True
        if self._name is None:
            self._name = name

        if self._ivar is not NULL:
            if self._ivar is None:
                ivname = "_" + self._name
            else:
                ivname = self._ivar

            if self.__parent is None:
                ivar_ref = ivar(name=ivname, type=self._typestr)
                class_dict[ivname] = ivar_ref

        if self._ro:
            self._setter = None

        else:
            setterName = (
                b"set"
                + name[0].upper().encode("latin1")
                + name[1:].encode("latin1")
                + b":"
            )
            signature = b"v@:" + self._typestr
            if self._setter is None:
                if self.__inherit:
                    pass

                elif self._dynamic:
                    dynSetterName = "set" + name[0].upper() + name[1:] + "_"
                    self.__setprop = _dynamic_setter(dynSetterName)
                    instance_methods.add(setterName)

                else:

                    if self._ivar is NULL:
                        raise ValueError(
                            "Cannot create default setter for property " "without ivar"
                        )

                    setprop = selector(
                        attrsetter(self._name, ivname, self._copy),
                        selector=setterName,
                        signature=signature,
                    )
                    setprop.isHidden = True
                    instance_methods.add(setprop)

                    # Use dynamic setter to avoid problems when subclassing
                    self.__setprop = _dynamic_setter(_str(setterName))
            else:
                setprop = selector(
                    self._setter, selector=setterName, signature=signature
                )
                setprop.isHidden = True
                instance_methods.add(setprop)

                # Use dynamic setter to avoid problems when subclassing
                self.__setprop = _dynamic_setter(_str(setterName))

        if self._typestr in (_C_NSBOOL, _C_BOOL):
            getterName = (
                b"is" + name[0].upper().encode("latin1") + name[1:].encode("latin1")
            )
        else:
            getterName = self._name.encode("latin1")

        if self._getter is None:
            if self.__inherit:
                pass

            elif self._dynamic:
                if self._typestr in (_C_NSBOOL, _C_BOOL):
                    dynGetterName = "is" + name[0].upper() + name[1:]
                else:
                    dynGetterName = self._name

                self.__getprop = _dynamic_getter(dynGetterName)
                instance_methods.add(getterName)

            else:
                if self._ivar is NULL:
                    raise ValueError(
                        "Cannot create default getter for property without ivar"
                    )

                self.__getprop = selector(
                    attrgetter(ivname),
                    selector=getterName,
                    signature=self._typestr + b"@:",
                )
                self.__getprop.isHidden = True
                instance_methods.add(self.__getprop)

        else:
            self.__getprop = getprop = selector(
                self._getter, selector=getterName, signature=self._typestr + b"@:"
            )
            getprop.isHidden = True
            instance_methods.add(getprop)
            # self.__getprop = _dynamic_getter(getterName)

        if self._validate is not None:
            selName = (
                b"validate"
                + self._name[0].upper().encode("latin")
                + self._name[1:].encode("latin")
                + b":error:"
            )
            signature = _C_NSBOOL + b"@:N^@o^@"
            validate = selector(self._validate, selector=selName, signature=signature)
            class_dict[validate.selector] = validate
            instance_methods.add(validate)

        if self._depends_on:
            if self.__parent is not None:
                if self.__parent._depends_on:
                    self._depends_on.update(self.__parent._depends_on.copy())

            self._depends_on = self._depends_on

            affecting = selector(
                _return_value(NSSet.setWithArray_(list(self._depends_on))),
                selector=b"keyPathsForValuesAffecting"
                + self._name[0].upper().encode("latin1")
                + self._name[1:].encode("latin1"),
                signature=b"@@:",
                isClassMethod=True,
            )
            class_dict[affecting.selector] = affecting
            class_methods.add(affecting)

    def __get__(self, an_object, owner):
        if an_object is None:
            return self
        return self.__getprop(an_object)

    def __set__(self, an_object, value):
        if self.__setprop is None:
            raise ValueError("setting read-only property " + self._name)

        return self.__setprop(an_object, value)

    def __delete__(self, an_object):
        raise TypeError("cannot delete property " + self._name)

    def depends_on(self, keypath):
        if self._depends_on is None:
            self._depends_on = set()
        self._depends_on.add(keypath)

    def getter(self, function):
        if self.__created:
            v = self._clone()
            v._getter = function
            return v

        self._getter = function
        return self

    def setter(self, function):

        if self.__created:
            v = self._clone()
            v._ro = False
            v._setter = function
            return v

        if self._ro:
            raise ValueError("Defining settter for read-only property")

        self._setter = function
        return self

    def validate(self, function):
        if self._ro:
            raise ValueError("Defining validator for read-only property")

        if self.__created:
            v = self._clone()
            v._validate = function
            return v

        self._validate = function
        return self


class bool_property(object_property):
    def __init__(
        self,
        name=None,
        read_only=False,
        copy=False,
        dynamic=False,
        ivar=None,
        typestr=_C_NSBOOL,
    ):
        super(bool_property, self).__init__(
            name, read_only, copy, dynamic, ivar, typestr
        )


NSIndexSet = lookUpClass("NSIndexSet")
NSMutableIndexSet = lookUpClass("NSMutableIndexSet")
NSKeyValueChangeSetting = 1
NSKeyValueChangeInsertion = 2
NSKeyValueChangeRemoval = 3
NSKeyValueChangeReplacement = 4


# Helper function for (not) pickling array_proxy instances
# NOTE: Don't remove this function, it can be referenced from
# pickle files.
def _id(value):
    return value


class array_proxy(collections.abc.MutableSequence):
    __slots__ = ("_name", "_parent", "__wrapped", "_ro")

    def __init__(self, name, parent, wrapped, read_only):
        self._name = name
        self._parent = parent
        self._ro = read_only
        self.__wrapped = wrapped

    @property
    def _wrapped(self):
        return self.__wrapped.__getvalue__(self._parent)

    def __indexSetForIndex(self, index):
        if isinstance(index, slice):
            result = NSMutableIndexSet.alloc().init()
            start, stop, step = index.indices(len(self._wrapped))
            for i in range(start, stop, step):
                result.addIndex_(i)

            return result

        elif isinstance(index, int):
            if index < 0:
                v = len(self) + index
                if v < 0:
                    raise IndexError(index)
                return NSIndexSet.alloc().initWithIndex_(v)

            else:
                return NSIndexSet.alloc().initWithIndex_(index)

        else:
            raise TypeError(index)

    def __repr__(self):
        return (
            "<array proxy for property " + self._name + " " + repr(self._wrapped) + ">"
        )

    def __reduce__(self):
        # Ensure that the proxy itself doesn't get stored
        # in pickles.
        return _id, (self._wrapped,)

    def __getattr__(self, name):
        # Default: just defer to wrapped list
        return getattr(self._wrapped, name)

    def __len__(self):
        return self._wrapped.__len__()

    def __getitem__(self, index):
        return self._wrapped[index]

    def __setitem__(self, index, value):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        indexes = self.__indexSetForIndex(index)
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeSetting, indexes, self._name
        )
        try:
            self._wrapped[index] = value
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeReplacement, indexes, self._name
            )

    def __delitem__(self, index):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        indexes = self.__indexSetForIndex(index)
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeRemoval, indexes, self._name
        )
        try:
            del self._wrapped[index]
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeRemoval, indexes, self._name
            )

    def append(self, value):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        index = len(self)
        indexes = NSIndexSet.alloc().initWithIndex_(index)
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeInsertion, indexes, self._name
        )
        try:
            self._wrapped.append(value)
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeInsertion, indexes, self._name
            )

    def insert(self, index, value):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        if isinstance(index, slice):
            raise TypeError("insert argument 1 is a slice")

        indexes = self.__indexSetForIndex(index)
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeInsertion, indexes, self._name
        )
        try:
            self._wrapped.insert(index, value)
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeInsertion, indexes, self._name
            )

    def pop(self, index=-1):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        if isinstance(index, slice):
            raise TypeError("insert argument 1 is a slice")

        indexes = self.__indexSetForIndex(index)
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeRemoval, indexes, self._name
        )
        try:
            return self._wrapped.pop(index)
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeRemoval, indexes, self._name
            )

    def extend(self, values):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        values = list(values)

        indexes = NSIndexSet.alloc().initWithIndexesInRange_((len(self), len(values)))

        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeInsertion, indexes, self._name
        )
        try:
            self._wrapped.extend(values)

        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeInsertion, indexes, self._name
            )

    def __iadd__(self, values):
        self.extend(values)
        return self

    def __add__(self, values):
        return self._wrapped + values

    def __mul__(self, count):
        return self._wrapped * count

    def __imul__(self, count):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))
        if not isinstance(count, int):
            raise TypeError(count)

        indexes = NSIndexSet.alloc().initWithIndexesInRange_(
            (len(self), len(self) * (count - 1))
        )
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeInsertion, indexes, self._name
        )
        try:
            self._wrapped.__imul__(count)
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeInsertion, indexes, self._name
            )

        return self

    def __eq__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped == other._wrapped

        else:
            return self._wrapped == other

    def __ne__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped != other._wrapped

        else:
            return self._wrapped != other

    def __lt__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped < other._wrapped

        else:
            return self._wrapped < other

    def __le__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped <= other._wrapped

        else:
            return self._wrapped <= other

    def __gt__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped > other._wrapped

        else:
            return self._wrapped > other

    def __ge__(self, other):
        if isinstance(other, array_proxy):
            return self._wrapped >= other._wrapped

        else:
            return self._wrapped >= other

    def sort(self, key=None, reverse=False):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        indexes = NSIndexSet.alloc().initWithIndexesInRange_((0, len(self._wrapped)))
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeReplacement, indexes, self._name
        )
        try:
            self._wrapped.sort(key=key, reverse=reverse)
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeReplacement, indexes, self._name
            )

    def reverse(self):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        indexes = NSIndexSet.alloc().initWithIndexesInRange_((0, len(self._wrapped)))
        self._parent.willChange_valuesAtIndexes_forKey_(
            NSKeyValueChangeReplacement, indexes, self._name
        )
        try:
            self._wrapped.reverse()
        finally:
            self._parent.didChange_valuesAtIndexes_forKey_(
                NSKeyValueChangeReplacement, indexes, self._name
            )


def makeArrayAccessors(name):
    def countOf(self):
        return len(getattr(self, name))

    def objectIn(self, idx):
        return getattr(self, name)[idx]

    def insert(self, value, idx):
        getattr(self, name).insert(idx, value)

    def replace(self, idx, value):
        getattr(self, name)[idx] = value

    def remove(self, idx):
        del getattr(self, name)[idx]

    return countOf, objectIn, insert, remove, replace


class array_property(object_property):
    def __init__(
        self,
        name=None,
        read_only=False,
        copy=True,
        dynamic=False,
        ivar=None,
        depends_on=None,
    ):
        super(array_property, self).__init__(
            name,
            read_only=read_only,
            copy=copy,
            dynamic=dynamic,
            ivar=ivar,
            depends_on=depends_on,
        )

    def __pyobjc_class_setup__(self, name, class_dict, instance_methods, class_methods):
        super(array_property, self).__pyobjc_class_setup__(
            name, class_dict, instance_methods, class_methods
        )

        # Insert (Mutable) Indexed Accessors

        name = self._name
        Name = name[0].upper() + name[1:]

        countOf, objectIn, insert, remove, replace = makeArrayAccessors(self._name)

        countOf = selector(
            countOf,
            selector=("countOf%s" % (Name,)).encode("latin1"),
            signature=_C_NSUInteger + b"@:",
        )
        countOf.isHidden = True
        instance_methods.add(countOf)

        objectIn = selector(
            objectIn,
            selector=("objectIn%sAtIndex:" % (Name,)).encode("latin1"),
            signature=b"@@:" + _C_NSUInteger,
        )
        objectIn.isHidden = True
        instance_methods.add(objectIn)

        insert = selector(
            insert,
            selector=("insertObject:in%sAtIndex:" % (Name,)).encode("latin1"),
            signature=b"v@:@" + _C_NSUInteger,
        )
        insert.isHidden = True
        instance_methods.add(insert)

        remove = selector(
            remove,
            selector=("removeObjectFrom%sAtIndex:" % (Name,)).encode("latin1"),
            signature=b"v@:" + _C_NSUInteger,
        )
        remove.isHidden = True
        instance_methods.add(remove)

        replace = selector(
            replace,
            selector=("replaceObjectIn%sAtIndex:withObject:" % (Name,)).encode(
                "latin1"
            ),
            signature=b"v@:" + _C_NSUInteger + b"@",
        )
        replace.isHidden = True
        instance_methods.add(replace)

    def __set__(self, an_object, value):
        if isinstance(value, array_proxy):
            if value._name == self._name and value._parent is an_object:
                # attr.prop = attr.prop
                return

        if isinstance(value, array_proxy):
            value = list(value)

        super(array_property, self).__set__(an_object, value)

    def __get__(self, an_object, owner):
        v = object_property.__get__(self, an_object, owner)
        if v is None:
            v = []
            object_property.__set__(self, an_object, v)
        return array_proxy(self._name, an_object, self, self._ro)

    def __getvalue__(self, an_object):
        v = object_property.__get__(self, an_object, None)
        if v is None:
            v = []
            object_property.__set__(self, an_object, v)
        return v


NSKeyValueUnionSetMutation = 1
NSKeyValueMinusSetMutation = 2
NSKeyValueIntersectSetMutation = 3
NSKeyValueSetSetMutation = 4


class set_proxy(collections.abc.MutableSet):
    __slots__ = ("_name", "__wrapped", "_parent", "_ro")

    def __init__(self, name, parent, wrapped, read_only):
        self._name = name
        self._parent = parent
        self._ro = read_only
        self.__wrapped = wrapped

    def __repr__(self):
        return "<set proxy for property " + self._name + " " + repr(self._wrapped) + ">"

    @property
    def _wrapped(self):
        return self.__wrapped.__getvalue__(self._parent)

    def __reduce__(self):
        # Ensure that the proxy itself doesn't get stored
        # in pickles.
        return _id, (self._wrapped,)

    def __getattr__(self, attr):
        return getattr(self._wrapped, attr)

    def __contains__(self, value):
        return self._wrapped.__contains__(value)

    def __iter__(self):
        return self._wrapped.__iter__()

    def __len__(self):
        return self._wrapped.__len__()

    def __eq__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped == other._wrapped

        else:
            return self._wrapped == other

    def __ne__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped != other._wrapped

        else:
            return self._wrapped != other

    def __lt__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped < other._wrapped

        else:
            return self._wrapped < other

    def __le__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped <= other._wrapped

        else:
            return self._wrapped <= other

    def __gt__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped > other._wrapped

        else:
            return self._wrapped > other

    def __ge__(self, other):
        if isinstance(other, set_proxy):
            return self._wrapped >= other._wrapped

        else:
            return self._wrapped >= other

    if sys.version_info[0] == 2:  # pragma: no 3.x cover; pragma: no branch

        def __cmp__(self, other):
            raise TypeError("cannot compare sets using cmp()")

    def add(self, item):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueUnionSetMutation, {item}
        )
        try:
            self._wrapped.add(item)
        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueUnionSetMutation, {item}
            )

    def clear(self):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        current_value = set(self._wrapped)
        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueMinusSetMutation, current_value
        )
        try:
            self._wrapped.clear()
        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueMinusSetMutation, current_value
            )

    def difference_update(self, *others):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        s = set()
        s.update(*others)
        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueMinusSetMutation, s
        )
        try:
            self._wrapped.difference_update(s)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueMinusSetMutation, s
            )

    def discard(self, item):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueMinusSetMutation, {item}
        )
        try:
            self._wrapped.discard(item)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueMinusSetMutation, {item}
            )

    def intersection_update(self, other):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        other = set(other)

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueIntersectSetMutation, other
        )
        try:
            self._wrapped.intersection_update(other)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueIntersectSetMutation, other
            )

    def pop(self):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        try:
            v = next(iter(self))
        except StopIteration:
            raise KeyError("Empty set")

        self.remove(v)
        return v

    def remove(self, item):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueMinusSetMutation, {item}
        )
        try:
            self._wrapped.remove(item)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueMinusSetMutation, {item}
            )

    def symmetric_difference_update(self, other):
        # NOTE: This method does not call the corresponding method
        # of the wrapped set to ensure that we generate the right
        # notifications.
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        other = set(other)

        to_add = set()
        to_remove = set()
        for o in other:
            if o in self:
                to_remove.add(o)
            else:
                to_add.add(o)

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueMinusSetMutation, to_remove
        )
        try:
            self._wrapped.difference_update(to_remove)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueMinusSetMutation, to_remove
            )

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueUnionSetMutation, to_add
        )
        try:
            self._wrapped.update(to_add)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueUnionSetMutation, to_add
            )

    def update(self, *others):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        s = set()
        s.update(*others)

        self._parent.willChangeValueForKey_withSetMutation_usingObjects_(
            self._name, NSKeyValueUnionSetMutation, s
        )
        try:
            self._wrapped.update(s)

        finally:
            self._parent.didChangeValueForKey_withSetMutation_usingObjects_(
                self._name, NSKeyValueUnionSetMutation, s
            )

    def __or__(self, other):
        return self._wrapped | other

    def __and__(self, other):
        return self._wrapped & other

    def __xor__(self, other):
        return self._wrapped ^ other

    def __sub__(self, other):
        return self._wrapped - other

    def __ior__(self, other):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self.update(other)
        return self

    def __isub__(self, other):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self.difference_update(other)
        return self

    def __ixor__(self, other):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self.symmetric_difference_update(other)
        return self

    def __iand__(self, other):
        if self._ro:
            raise ValueError("Property '%s' is read-only" % (self._name,))

        self.intersection_update(other)
        return self


def makeSetAccessors(name):
    def countOf(self):
        return len(getattr(self, name))

    def enumeratorOf(self):
        return iter(getattr(self, name))

    def memberOf(self, value):
        collection = getattr(self, name)

        for item in collection:
            if item == value:
                return item

        else:
            return None

    def add(self, value):
        getattr(self, name).add(value)

    def remove(self, value):
        getattr(self, name).discard(value)

    return countOf, enumeratorOf, memberOf, add, remove


class set_property(object_property):
    def __init__(
        self,
        name=None,
        read_only=False,
        copy=True,
        dynamic=False,
        ivar=None,
        depends_on=None,
    ):
        super(set_property, self).__init__(
            name,
            read_only=read_only,
            copy=copy,
            dynamic=dynamic,
            ivar=ivar,
            depends_on=depends_on,
        )

    def __get__(self, an_object, owner):
        v = object_property.__get__(self, an_object, owner)
        if v is None:
            v = set()
            object_property.__set__(self, an_object, v)
        return set_proxy(self._name, an_object, self, self._ro)

    def __set__(self, an_object, value):
        if isinstance(value, set_proxy):
            if value._name == self._name and value._parent is an_object:
                # attr.prop = attr.prop
                return

        if isinstance(value, set_proxy):
            value = list(value)

        super(set_property, self).__set__(an_object, value)

    def __getvalue__(self, an_object):
        v = object_property.__get__(self, an_object, None)
        if v is None:
            v = set()
            object_property.__set__(self, an_object, v)
        return v

    def __pyobjc_class_setup__(self, name, class_dict, instance_methods, class_methods):
        super(set_property, self).__pyobjc_class_setup__(
            name, class_dict, instance_methods, class_methods
        )

        # (Mutable) Unordered Accessors

        name = self._name
        Name = name[0].upper() + name[1:]

        countOf, enumeratorOf, memberOf, add, remove = makeSetAccessors(self._name)

        countOf = selector(
            countOf,
            selector=("countOf%s" % (Name,)).encode("latin1"),
            signature=_C_NSUInteger + b"@:",
        )
        countOf.isHidden = True
        instance_methods.add(countOf)

        enumeratorOf = selector(
            enumeratorOf,
            selector=("enumeratorOf%s" % (Name,)).encode("latin1"),
            signature=b"@@:",
        )
        enumeratorOf.isHidden = True
        instance_methods.add(enumeratorOf)

        memberOf = selector(
            memberOf,
            selector=("memberOf%s:" % (Name,)).encode("latin"),
            signature=b"@@:@",
        )
        memberOf.isHidden = True
        instance_methods.add(memberOf)

        add1 = selector(
            add, selector=("add%s:" % (Name,)).encode("latin"), signature=b"v@:@"
        )
        add1.isHidden = True
        instance_methods.add(add1)

        add2 = selector(
            add, selector=("add%sObject:" % (Name,)).encode("latin1"), signature=b"v@:@"
        )
        add2.isHidden = True
        instance_methods.add(add2)

        remove1 = selector(
            remove, selector=("remove%s:" % (Name,)).encode("latin1"), signature=b"v@:@"
        )
        remove1.isHidden = True
        instance_methods.add(remove1)

        remove2 = selector(
            remove,
            selector=("remove%sObject:" % (Name,)).encode("latin"),
            signature=b"v@:@",
        )
        remove2.isHidden = True
        instance_methods.add(remove2)


NSMutableDictionary = lookUpClass("NSMutableDictionary")


class dict_property(object_property):
    def __get__(self, an_object, owner):
        v = object_property.__get__(self, an_object, owner)
        if v is None:
            v = NSMutableDictionary.alloc().init()
            object_property.__set__(self, an_object, v)
        return object_property.__get__(self, an_object, owner)



"""
Convenience interface for NSSet/NSMutableSet
"""
__all__ = ()

import collections.abc

from objc._convenience import addConvenienceForClass, container_unwrap, container_wrap
from objc._objc import lookUpClass

NSSet = lookUpClass("NSSet")
NSMutableSet = lookUpClass("NSMutableSet")

collections.abc.Set.register(NSSet)
collections.abc.MutableSet.register(NSMutableSet)


def nsset_isdisjoint(self, other):
    if not hasattr(other, "__contains__"):
        other = list(other)

    for item in self:
        if item in other:
            return False
    return True


def nsset_union(self, *other):
    result = NSMutableSet()
    result.unionSet_(self)
    for val in other:
        if isinstance(val, collections.abc.Set):
            result.unionSet_(val)
        else:
            result.unionSet_(set(val))
    return result


def nsset_intersection(self, *others):
    if len(others) == 0:
        return self.mutableCopy()

    others = [o if hasattr(o, "__contains__") else list(o) for o in others]

    result = NSMutableSet()
    for item in self:
        for o in others:
            if item not in o:
                break
        else:
            result.add(item)
    return result


def nsset_difference(self, *others):
    result = self.mutableCopy()

    for value in others:
        if isinstance(value, collections.abc.Set):
            result.minusSet_(value)
        else:
            result.minusSet_(set(value))

    return result


def nsset_symmetric_difference(self, other):
    result = NSMutableSet()
    if not hasattr(other, "__contains__"):
        other = list(other)
    for item in iter(self):
        if item not in other:
            result.add(item)
    for item in other:
        if item not in self:
            result.add(item)
    return result


def nsset__contains__(self, value):
    hash(value)  # Force error for non-hashable values
    return self.containsObject_(value)


def nsset__or__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("NSSet|value where value is not a set")
    return nsset_union(self, other)


def nsset__ror__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("value|NSSet where value is not a set")
    return nsset_union(other, self)


def nsset__and__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("NSSet&value where value is not a set")
    return nsset_intersection(self, other)


def nsset__rand__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("value&NSSet where value is not a set")
    return nsset_intersection(other, self)


def nsset__sub__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("NSSet-value where value is not a set")
    return nsset_difference(self, other)


def nsset__rsub__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("NSSet-value where value is not a set")
    return nsset_difference(NSMutableSet(other), self)


def nsset__xor__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError("NSSet-value where value is not a set")
    return nsset_symmetric_difference(other, self)


def nsset_issubset(self, other):
    if isinstance(other, collections.abc.Set):
        return self.isSubsetOfSet_(other)

    else:
        return self.isSubsetOfSet_(set(other))


def nsset__le__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError()
    return nsset_issubset(self, other)


def nsset__eq__(self, other):
    if not isinstance(other, collections.abc.Set):
        return False

    return self.isEqualToSet_(other)


def nsset__ne__(self, other):
    if not isinstance(other, collections.abc.Set):
        return True

    return not self.isEqualToSet_(other)


def nsset__lt__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError()

    return (self <= other) and (self != other)


def nsset_issuperset(self, other):
    if not isinstance(other, collections.abc.Set):
        other = set(other)

    for item in other:
        if item not in self:
            return False

    return True


def nsset__ge__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError()
    return nsset_issuperset(self, other)


def nsset__gt__(self, other):
    if not isinstance(other, collections.abc.Set):
        raise TypeError()
    return (self >= other) and (self != other)


def nsset_update(self, *others):
    for other in others:
        if isinstance(other, collections.abc.Set):
            self.unionSet_(other)
        else:
            self.unionSet_(set(other))


def nsset_intersection_update(self, *others):
    for other in others:
        if isinstance(other, collections.abc.Set):
            self.intersectSet_(other)
        else:
            self.intersectSet_(set(other))


def nsset_difference_update(self, *others):
    for other in others:
        if isinstance(other, collections.abc.Set):
            self.minusSet_(other)
        else:
            self.minusSet_(set(other))


def nsset_symmetric_difference_update(self, other):
    toadd = set()
    toremove = set()

    if isinstance(other, collections.abc.Set):
        totest = other
    else:
        totest = set(other)

    for value in self:
        if value in totest:
            toremove.add(value)
    for value in totest:
        if value not in self:
            toadd.add(value)

    self.minusSet_(toremove)
    self.unionSet_(toadd)


def nsset_pop(self):
    if len(self) == 0:
        raise KeyError()

    v = self.anyObject()
    self.removeObject_(v)
    return container_unwrap(v, KeyError)


def nsset_remove(self, value):
    hash(value)
    value = container_wrap(value)
    if value not in self:
        raise KeyError(value)
    self.removeObject_(value)


def nsset_discard(self, value):
    hash(value)
    self.removeObject_(container_wrap(value))


def nsset_add(self, value):
    hash(value)
    self.addObject_(container_wrap(value))


class nsset__iter__(object):
    def __init__(self, value):
        self._size = len(value)
        self._enum = value.objectEnumerator()

    def __length_hint__(self):
        return self._size

    def __iter__(self):
        return self

    def __next__(self):
        self._size -= 1
        return container_unwrap(self._enum.nextObject(), StopIteration)


addConvenienceForClass(
    "NSSet",
    (
        ("__len__", lambda self: self.count()),
        ("__iter__", lambda self: nsset__iter__(self)),
        ("__contains__", nsset__contains__),
        ("isdisjoint", nsset_isdisjoint),
        ("union", nsset_union),
        ("intersection", nsset_intersection),
        ("difference", nsset_difference),
        ("symmetric_difference", nsset_symmetric_difference),
        ("issubset", nsset_issubset),
        ("__eq__", nsset__eq__),
        ("__ne__", nsset__ne__),
        ("__le__", nsset__le__),
        ("__lt__", nsset__lt__),
        ("issuperset", nsset_issuperset),
        ("__ge__", nsset__ge__),
        ("__gt__", nsset__gt__),
        ("__or__", nsset__or__),
        ("__ror__", nsset__ror__),
        ("__and__", nsset__and__),
        ("__rand__", nsset__rand__),
        ("__xor__", nsset__xor__),
        ("__rxor__", nsset__xor__),
        ("__sub__", nsset__sub__),
        ("__rsub__", nsset__rsub__),
    ),
)

addConvenienceForClass(
    "NSMutableSet",
    (
        ("add", nsset_add),
        ("remove", nsset_remove),
        ("discard", nsset_discard),
        ("update", nsset_update),
        ("intersection_update", nsset_intersection_update),
        ("difference_update", nsset_difference_update),
        ("symmetric_difference_update", nsset_symmetric_difference_update),
        ("clear", lambda self: self.removeAllObjects()),
        ("pop", nsset_pop),
    ),
)


def nsset_new(cls, sequence=None):
    if not sequence:
        return NSSet.set()

    if isinstance(sequence, (NSSet, set, frozenset)):
        return NSSet.set().setByAddingObjectsFromSet_(sequence)

    else:
        return NSSet.set().setByAddingObjectsFromSet_(set(sequence))


def nsmutableset_new(cls, sequence=None):
    if not sequence:
        value = NSMutableSet.set()

    elif isinstance(sequence, (NSSet, set, frozenset)):
        value = NSMutableSet.set()
        value.unionSet_(sequence)

    else:
        value = NSMutableSet.set()
        value.unionSet_(set(sequence))

    return value


addConvenienceForClass("NSSet", (("__new__", staticmethod(nsset_new)),))

addConvenienceForClass("NSMutableSet", (("__new__", staticmethod(nsmutableset_new)),))



"""
This module implements a callback function that is used by the C code to
add Python special methods to Objective-C classes with a suitable interface.
"""
from objc._convenience import addConvenienceForClass

__all__ = ("addConvenienceForBasicSequence",)


def seq_iter(self):
    for idx in range(len(self)):
        yield self[idx]


def seq_len(self):
    return self.count()


def seq_getitem(self, idx):
    if isinstance(idx, slice):
        raise ValueError("Slices not supported")

    if idx < 0:
        orig_idx = idx
        idx += seq_len(self)

        if idx < 0:
            raise IndexError(orig_idx)

    return self.objectAtIndex_(idx)


def seq_setitem(self, idx, value):
    if isinstance(idx, slice):
        raise ValueError("Slices not supported")

    if idx < 0:
        orig_idx = idx
        idx += seq_len(self)

        if idx < 0:
            raise IndexError(orig_idx)

    return self.setObject_atIndex_(value, idx)


def addConvenienceForBasicSequence(classname, readonly=True):
    addConvenienceForClass(
        classname,
        (("__len__", seq_len), ("__getitem__", seq_getitem), ("__iter__", seq_iter)),
    )

    if not readonly:
        addConvenienceForClass(classname, (("__setitem__", seq_setitem),))



"""
Python <-> Objective-C bridge (PyObjC)

This module defines the core interfaces of the Python<->Objective-C bridge.
"""
from . import _objc


# Import the namespace from the _objc extension
def _update(g):
    for k in _objc.__dict__:
        g.setdefault(k, getattr(_objc, k))


_update(globals())
del _update


from ._convenience import *  # noqa: F401, F403, E402
from ._convenience_nsobject import *  # noqa: F401, F403, E402
from ._convenience_nsdecimal import *  # noqa: F401, F403, E402
from ._convenience_nsdata import *  # noqa: F401, F403, E402
from ._convenience_nsdictionary import *  # noqa: F401, F403, E402
from ._convenience_nsset import *  # noqa: F401, F403, E402
from ._convenience_nsarray import *  # noqa: F401, F403, E402
from ._convenience_nsstring import *  # noqa: F401, F403, E402
from ._convenience_mapping import *  # noqa: F401, F403, E402
from ._convenience_sequence import *  # noqa: F401, F403, E402
from ._bridgesupport import *  # noqa: F401, F403, E402
from ._dyld import *  # noqa: F401, F403, E402
from ._protocols import *  # noqa: F401, F403, E402
from ._descriptors import *  # noqa: F401, F403, E402
from ._category import *  # noqa: F401, F403, E402
from ._bridges import *  # noqa: F401, F403, E402
from ._pythonify import *  # noqa: F401, F403, E402
from ._locking import *  # noqa: F401, F403, E402
from ._context import *  # noqa: F401, F403, E402
from ._properties import *  # noqa: F401, F403, E402
from ._lazyimport import *  # noqa: F401, F403, E402
from ._compat import *  # noqa: F401, F403, E402
from . import _callable_docstr  # noqa: F401, F403, E402
from . import _pycoder  # noqa: F401, F403, E402


# Helper function for new-style metadata modules
def _resolve_name(name):
    if "." not in name:
        raise ValueError(name)

    module, name = name.rsplit(".", 1)
    m = __import__(module)
    for k in module.split(".")[1:]:
        m = getattr(m, k)

    return getattr(m, name)


_NSAutoreleasePool = None

# Aliases for some common Objective-C constants
nil = None
YES = True
NO = False


class autorelease_pool(object):
    """
    A context manager that runs the body of the block with a fresh
    autorelease pool. The actual release pool is not accessible.
    """

    def __init__(self):
        global _NSAutoreleasePool
        if _NSAutoreleasePool is None:
            _NSAutoreleasePool = lookUpClass("NSAutoreleasePool")  # noqa: F405

    def __enter__(self):
        self._pool = _NSAutoreleasePool.alloc().init()

    def __exit__(self, exc_type, value, tp):
        del self._pool
